from typing import TYPE_CHECKING

import commands2
from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *

import pytest


def test_waitCommand(scheduler: commands2.CommandScheduler):
    with ManualSimTime() as sim:
        waitCommand = commands2.WaitCommand(2)

        scheduler.schedule(waitCommand)
        scheduler.run()
        sim.step(1)
        scheduler.run()

        assert scheduler.isScheduled(waitCommand)

        sim.step(2)

        scheduler.run()

        assert not scheduler.isScheduled(waitCommand)


def test_withTimeout(scheduler: commands2.CommandScheduler):
    with ManualSimTime() as sim:
        command1 = commands2.Command()
        start_spying_on(command1)

        timeout = command1.withTimeout(2)

        scheduler.schedule(timeout)
        scheduler.run()

        verify(command1).initialize()
        verify(command1).execute()
        assert not scheduler.isScheduled(command1)
        assert scheduler.isScheduled(timeout)

        sim.step(3)
        scheduler.run()

        verify(command1).end(True)
        verify(command1, never()).end(False)
        assert not scheduler.isScheduled(timeout)



# validated: 2024-01-20 DS 3ba501f9478a button/CommandXboxController.java
from typing import Optional

from wpilib import XboxController
from wpilib.event import EventLoop

from ..commandscheduler import CommandScheduler
from .commandgenerichid import CommandGenericHID
from .trigger import Trigger


class CommandXboxController(CommandGenericHID):
    """
    A version of XboxController with Trigger factories for command-based.
    """

    _hid: XboxController

    def __init__(self, port: int):
        """
        Construct an instance of a controller.

        :param port: The port index on the Driver Station that the controller is plugged into.
        """
        super().__init__(port)
        self._hid = XboxController(port)

    def getHID(self) -> XboxController:
        """
        Get the underlying GenericHID object.

        :returns: the wrapped GenericHID object
        """
        return self._hid

    def leftBumper(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the left bumper's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the right bumper's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getLeftBumper())

    def rightBumper(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the right bumper's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the left bumper's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getRightBumper())

    def leftStick(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the left stick button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the left stick button's digital signal attached to the
                  given loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getLeftStickButton())

    def rightStick(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the right stick button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the right stick button's digital signal attached to the
                  given loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getRightStickButton())

    def a(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the A button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the A button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getAButton())

    def b(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the B button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the B button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getBButton())

    def x(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the X button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the X button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getXButton())

    def y(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the Y button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the Y button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getYButton())

    def start(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the start button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the start button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getStartButton())

    def back(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the back button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the back button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getBackButton())

    def leftTrigger(
        self, threshold: float = 0.5, loop: Optional[EventLoop] = None
    ) -> Trigger:
        """
        Constructs a Trigger instance around the axis value of the left trigger. The returned trigger
        will be true when the axis value is greater than {@code threshold}.

        :param threshold: the minimum axis value for the returned Trigger to be true. This value
                          should be in the range [0, 1] where 0 is the unpressed state of the axis.
        :param loop: the event loop instance to attach the Trigger to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: a Trigger instance that is true when the left trigger's axis exceeds the provided
            threshold, attached to the given event loop
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getLeftTriggerAxis() > threshold)

    def rightTrigger(
        self, threshold: float = 0.5, loop: Optional[EventLoop] = None
    ) -> Trigger:
        """
        Constructs a Trigger instance around the axis value of the right trigger. The returned trigger
        will be true when the axis value is greater than {@code threshold}.

        :param threshold: the minimum axis value for the returned Trigger to be true. This value
                          should be in the range [0, 1] where 0 is the unpressed state of the axis.
        :param loop: the event loop instance to attach the Trigger to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: a Trigger instance that is true when the right trigger's axis exceeds the provided
                  threshold, attached to the given event loop
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getRightTriggerAxis() > threshold)

    def getLeftX(self) -> float:
        """
        Get the X axis value of left side of the controller.

        :returns: The axis value.
        """
        return self._hid.getLeftX()

    def getRightX(self) -> float:
        """
        Get the X axis value of right side of the controller.

        :returns: The axis value.
        """
        return self._hid.getRightX()

    def getLeftY(self) -> float:
        """
        Get the Y axis value of left side of the controller.

        :returns: The axis value.
        """
        return self._hid.getLeftY()

    def getRightY(self) -> float:
        """
        Get the Y axis value of right side of the controller.

        :returns: The axis value.
        """
        return self._hid.getRightY()

    def getLeftTriggerAxis(self) -> float:
        """
        Get the left trigger (LT) axis value of the controller. Note that this axis is bound to the
        range of [0, 1] as opposed to the usual [-1, 1].

        :returns: The axis value.
        """
        return self._hid.getLeftTriggerAxis()

    def getRightTriggerAxis(self) -> float:
        """
        Get the right trigger (RT) axis value of the controller. Note that this axis is bound to the
        range of [0, 1] as opposed to the usual [-1, 1].

        :returns: The axis value.
        """
        return self._hid.getRightTriggerAxis()



from typing import Generic, TypeVar

import commands2
import pytest

# T = TypeVar("T", bound=commands2.Command)
# T = commands2.Command

from util import *


class SingleCompositionTestBase:
    def composeSingle(self, member: commands2.Command):
        raise NotImplementedError

    @pytest.mark.parametrize(
        "interruptionBehavior",
        [
            commands2.InterruptionBehavior.kCancelSelf,
            commands2.InterruptionBehavior.kCancelIncoming,
        ],
    )
    def test_interruptible(self, interruptionBehavior: commands2.InterruptionBehavior):
        command = self.composeSingle(
            commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                interruptionBehavior
            )
        )
        assert command.getInterruptionBehavior() == interruptionBehavior

    @pytest.mark.parametrize("runsWhenDisabled", [True, False])
    def test_runWhenDisabled(self, runsWhenDisabled: bool):
        command = self.composeSingle(
            commands2.WaitUntilCommand(lambda: False).ignoringDisable(runsWhenDisabled)
        )
        assert command.runsWhenDisabled() == runsWhenDisabled

    def test_command_in_other_composition(self):
        command = commands2.InstantCommand()
        wrapped = commands2.WrapperCommand(command)
        with pytest.raises(commands2.IllegalCommandUse):
            self.composeSingle(command)

    def test_command_in_multiple_compositions(self):
        command = commands2.InstantCommand()
        self.composeSingle(command)
        with pytest.raises(commands2.IllegalCommandUse):
            self.composeSingle(command)

    def test_compose_then_schedule(self, scheduler: commands2.CommandScheduler):
        command = commands2.InstantCommand()
        self.composeSingle(command)
        with pytest.raises(commands2.IllegalCommandUse):
            scheduler.schedule(command)

    def test_schedule_then_compose(self, scheduler: commands2.CommandScheduler):
        command = commands2.RunCommand(lambda: None)
        scheduler.schedule(command)
        with pytest.raises(commands2.IllegalCommandUse):
            self.composeSingle(command)


class MultiCompositionTestBase(SingleCompositionTestBase):
    def compose(self, *members: commands2.Command):
        raise NotImplementedError

    def composeSingle(self, member: commands2.Command):
        return self.compose(member)

    @pytest.mark.parametrize(
        "expected,command1,command2,command3",
        [
            pytest.param(
                commands2.InterruptionBehavior.kCancelSelf,
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                id="AllCancelSelf",
            ),
            pytest.param(
                commands2.InterruptionBehavior.kCancelIncoming,
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                id="AllCancelIncoming",
            ),
            pytest.param(
                commands2.InterruptionBehavior.kCancelSelf,
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                id="TwoCancelSelfOneIncoming",
            ),
            pytest.param(
                commands2.InterruptionBehavior.kCancelSelf,
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelIncoming
                ),
                commands2.WaitUntilCommand(lambda: False).withInterruptBehavior(
                    commands2.InterruptionBehavior.kCancelSelf
                ),
                id="TwoCancelIncomingOneSelf",
            ),
        ],
    )
    def test_interruptible(self, expected, command1, command2, command3):
        command = self.compose(command1, command2, command3)
        assert command.getInterruptionBehavior() == expected

    @pytest.mark.parametrize(
        "expected,command1,command2,command3",
        [
            pytest.param(
                False,
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                id="AllFalse",
            ),
            pytest.param(
                True,
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                id="AllTrue",
            ),
            pytest.param(
                False,
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                id="TwoTrueOneFalse",
            ),
            pytest.param(
                False,
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(False),
                commands2.WaitUntilCommand(lambda: False).ignoringDisable(True),
                id="TwoFalseOneTrue",
            ),
        ],
    )
    def test_runWhenDisabled(self, expected, command1, command2, command3):
        command = self.compose(command1, command2, command3)
        assert command.runsWhenDisabled() == expected



# validated: 2024-01-20 DS ad0859a8c9ec WrapperCommand.java
from __future__ import annotations

from typing import Set

from .command import Command, InterruptionBehavior
from .commandscheduler import CommandScheduler


class WrapperCommand(Command):
    """
    A class used internally to wrap commands while overriding a specific method; all other methods
    will call through to the wrapped command.

    The rules for command compositions apply: command instances that are passed to it cannot be
    added to any other composition or scheduled individually, and the composition requires all
    subsystems its components require.
    """

    def __init__(self, command: Command):
        """
        Wrap a command.

        :param command: the command being wrapped. Trying to directly schedule this
                        command or add it to a composition will throw an exception.
        """
        super().__init__()

        CommandScheduler.getInstance().registerComposedCommands([command])
        self._command = command
        self.setName(self._command.getName())

    def initialize(self):
        """
        The initial subroutine of a command. Called once when the command is initially scheduled.
        """
        self._command.initialize()

    def execute(self):
        """
        The main body of a command. Called repeatedly while the command is scheduled.
        """
        self._command.execute()

    def end(self, interrupted: bool):
        """
        The action to take when the command ends. Called when either the command finishes normally, or
        when it interrupted/canceled.

        Do not schedule commands here that share requirements with this command. Use
        :func:`commands2.Command.andThen` instead.

        :param interrupted: whether the command was interrupted/canceled
        """
        self._command.end(interrupted)

    def isFinished(self) -> bool:
        """
        Whether the command has finished. Once a command finishes, the scheduler will call its end()
        method and un-schedule it.

        :returns: whether the command has finished.
        """
        return self._command.isFinished()

    def getRequirements(self) -> Set:
        """
        Specifies the set of subsystems used by this command. Two commands cannot use the same
        subsystem at the same time. If the command is scheduled as interruptible and another command is
        scheduled that shares a requirement, the command will be interrupted. Else, the command will
        not be scheduled. If no subsystems are required, return an empty set.

        Note: it is recommended that user implementations contain the requirements as a field, and
        return that field here, rather than allocating a new set every time this is called.

        :returns: the set of subsystems that are required
        """
        return self._command.getRequirements()

    def runsWhenDisabled(self) -> bool:
        """
        Whether the given command should run when the robot is disabled. Override to return true if the
        command should run when disabled.

        :returns: whether the command should run when the robot is disabled
        """
        return self._command.runsWhenDisabled()

    def getInterruptionBehavior(self) -> InterruptionBehavior:
        return self._command.getInterruptionBehavior()



from typing import TYPE_CHECKING

import commands2
from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *

import pytest
from wpilib import RobotState


def test_robotDisabledCommandCancel(scheduler: commands2.CommandScheduler):
    command = commands2.Command()
    scheduler.schedule(command)
    assert scheduler.isScheduled(command)
    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()
    scheduler.run()
    assert not scheduler.isScheduled(command)
    DriverStationSim.setEnabled(True)
    DriverStationSim.notifyNewData()


def test_runWhenDisabled(scheduler: commands2.CommandScheduler):
    command = commands2.Command()
    command.runsWhenDisabled = lambda: True

    scheduler.schedule(command)

    assert scheduler.isScheduled(command)

    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    scheduler.run()

    assert scheduler.isScheduled(command)


def test_sequentialGroupRunWhenDisabled(scheduler: commands2.CommandScheduler):
    command1 = commands2.Command()
    command1.runsWhenDisabled = lambda: True
    command2 = commands2.Command()
    command2.runsWhenDisabled = lambda: True
    command3 = commands2.Command()
    command3.runsWhenDisabled = lambda: True
    command4 = commands2.Command()
    command4.runsWhenDisabled = lambda: False

    runWhenDisabled = commands2.SequentialCommandGroup(command1, command2)
    dontRunWhenDisabled = commands2.SequentialCommandGroup(command3, command4)

    scheduler.schedule(runWhenDisabled)
    scheduler.schedule(dontRunWhenDisabled)

    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    scheduler.run()

    assert scheduler.isScheduled(runWhenDisabled)
    assert not scheduler.isScheduled(dontRunWhenDisabled)


def test_parallelGroupRunWhenDisabled(scheduler: commands2.CommandScheduler):
    command1 = commands2.Command()
    command1.runsWhenDisabled = lambda: True
    command2 = commands2.Command()
    command2.runsWhenDisabled = lambda: True
    command3 = commands2.Command()
    command3.runsWhenDisabled = lambda: True
    command4 = commands2.Command()
    command4.runsWhenDisabled = lambda: False

    runWhenDisabled = commands2.ParallelCommandGroup(command1, command2)
    dontRunWhenDisabled = commands2.ParallelCommandGroup(command3, command4)

    scheduler.schedule(runWhenDisabled)
    scheduler.schedule(dontRunWhenDisabled)

    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    scheduler.run()

    assert scheduler.isScheduled(runWhenDisabled)
    assert not scheduler.isScheduled(dontRunWhenDisabled)


def test_conditionalRunWhenDisabled(scheduler: commands2.CommandScheduler):
    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    command1 = commands2.Command()
    command1.runsWhenDisabled = lambda: True
    command2 = commands2.Command()
    command2.runsWhenDisabled = lambda: True
    command3 = commands2.Command()
    command3.runsWhenDisabled = lambda: True
    command4 = commands2.Command()
    command4.runsWhenDisabled = lambda: False

    runWhenDisabled = commands2.ConditionalCommand(command1, command2, lambda: True)
    dontRunWhenDisabled = commands2.ConditionalCommand(command3, command4, lambda: True)

    scheduler.schedule(runWhenDisabled, dontRunWhenDisabled)

    assert scheduler.isScheduled(runWhenDisabled)
    assert not scheduler.isScheduled(dontRunWhenDisabled)


def test_selectRunWhenDisabled(scheduler: commands2.CommandScheduler):
    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    command1 = commands2.Command()
    command1.runsWhenDisabled = lambda: True
    command2 = commands2.Command()
    command2.runsWhenDisabled = lambda: True
    command3 = commands2.Command()
    command3.runsWhenDisabled = lambda: True
    command4 = commands2.Command()
    command4.runsWhenDisabled = lambda: False

    runWhenDisabled = commands2.SelectCommand({1: command1, 2: command2}, lambda: 1)
    dontRunWhenDisabled = commands2.SelectCommand({1: command3, 2: command4}, lambda: 1)

    scheduler.schedule(runWhenDisabled, dontRunWhenDisabled)
    assert scheduler.isScheduled(runWhenDisabled)
    assert not scheduler.isScheduled(dontRunWhenDisabled)


def test_parallelConditionalRunWhenDisabledTest(scheduler: commands2.CommandScheduler):
    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()

    command1 = commands2.Command()
    command1.runsWhenDisabled = lambda: True
    command2 = commands2.Command()
    command2.runsWhenDisabled = lambda: True
    command3 = commands2.Command()
    command3.runsWhenDisabled = lambda: True
    command4 = commands2.Command()
    command4.runsWhenDisabled = lambda: False

    runWhenDisabled = commands2.ConditionalCommand(command1, command2, lambda: True)
    dontRunWhenDisabled = commands2.ConditionalCommand(command3, command4, lambda: True)

    parallel = commands2.cmd.parallel(runWhenDisabled, dontRunWhenDisabled)

    scheduler.schedule(parallel)

    assert not scheduler.isScheduled(runWhenDisabled)



# validated: 2024-01-20 DS aaea85ff1656 WaitUntilCommand.java
from __future__ import annotations

from typing import Callable, overload

from wpilib import Timer
from wpimath import units

from .command import Command
from .util import format_args_kwargs


class WaitUntilCommand(Command):
    """
    A command that does nothing but ends after a specified match time or condition. Useful for
    CommandGroups.
    """

    _condition: Callable[[], bool]

    @overload
    def __init__(self, condition: Callable[[], bool]):
        """
        Creates a new WaitUntilCommand that ends after a given condition becomes true.

        :param condition: the condition to determine when to end
        """
        ...

    @overload
    def __init__(self, time: units.seconds):
        """
        Creates a new WaitUntilCommand that ends after a given match time.

        NOTE: The match timer used for this command is UNOFFICIAL. Using this command does NOT
        guarantee that the time at which the action is performed will be judged to be legal by the
        referees. When in doubt, add a safety factor or time the action manually.

        :param time: the match time at which to end, in seconds
        """
        ...

    def __init__(self, *args, **kwargs):
        super().__init__()

        def init_condition(condition: Callable[[], bool]) -> None:
            assert callable(condition)
            self._condition = condition

        def init_time(time: float) -> None:
            self._condition = lambda: Timer.getMatchTime() - time > 0

        num_args = len(args) + len(kwargs)

        if num_args == 1 and len(kwargs) == 1:
            if "condition" in kwargs:
                return init_condition(kwargs["condition"])
            elif "time" in kwargs:
                return init_time(kwargs["time"])
        elif num_args == 1 and len(args) == 1:
            if isinstance(args[0], float):
                return init_time(args[0])
            elif callable(args[0]):
                return init_condition(args[0])

        raise TypeError(
            f"""
TypeError: WaitUntilCommand(): incompatible function arguments. The following argument types are supported:
    1. (self: WaitUntilCommand, condition: () -> bool)
    2. (self: WaitUntilCommand, time: wpimath.units.seconds)

Invoked with: {format_args_kwargs(self, *args, **kwargs)}
"""
        )

    def isFinished(self) -> bool:
        return self._condition()

    def runsWhenDisabled(self) -> bool:
        return True



from .command import Command, InterruptionBehavior

from . import button
from . import cmd

from .commandscheduler import CommandScheduler
from .conditionalcommand import ConditionalCommand
from .deferredcommand import DeferredCommand
from .exceptions import IllegalCommandUse
from .functionalcommand import FunctionalCommand
from .instantcommand import InstantCommand
from .notifiercommand import NotifierCommand
from .parallelcommandgroup import ParallelCommandGroup
from .paralleldeadlinegroup import ParallelDeadlineGroup
from .parallelracegroup import ParallelRaceGroup
from .pidcommand import PIDCommand
from .pidsubsystem import PIDSubsystem
from .printcommand import PrintCommand
from .profiledpidcommand import ProfiledPIDCommand
from .profiledpidsubsystem import ProfiledPIDSubsystem
from .proxycommand import ProxyCommand
from .repeatcommand import RepeatCommand
from .runcommand import RunCommand
from .schedulecommand import ScheduleCommand
from .selectcommand import SelectCommand
from .sequentialcommandgroup import SequentialCommandGroup
from .startendcommand import StartEndCommand
from .subsystem import Subsystem
from .swervecontrollercommand import SwerveControllerCommand
from .timedcommandrobot import TimedCommandRobot
from .trapezoidprofilecommand import TrapezoidProfileCommand
from .trapezoidprofilesubsystem import TrapezoidProfileSubsystem
from .waitcommand import WaitCommand
from .waituntilcommand import WaitUntilCommand
from .wrappercommand import WrapperCommand

from typing import TYPE_CHECKING

__all__ = [
    "button",
    "cmd",
    "Command",
    "CommandScheduler",
    "ConditionalCommand",
    "DeferredCommand",
    "FunctionalCommand",
    "IllegalCommandUse",
    "InstantCommand",
    "InterruptionBehavior",
    "NotifierCommand",
    "ParallelCommandGroup",
    "ParallelDeadlineGroup",
    "ParallelRaceGroup",
    "PIDCommand",
    "PIDSubsystem",
    "PrintCommand",
    "ProfiledPIDCommand",
    "ProfiledPIDSubsystem",
    "ProxyCommand",
    "RepeatCommand",
    "RunCommand",
    "ScheduleCommand",
    "SelectCommand",
    "SequentialCommandGroup",
    "StartEndCommand",
    "Subsystem",
    "SwerveControllerCommand",
    "TimedCommandRobot",
    "TrapezoidProfileCommand",
    "TrapezoidProfileSubsystem",
    "WaitCommand",
    "WaitUntilCommand",
    "WrapperCommand",
]

if not TYPE_CHECKING:

    def __getattr__(attr):
        if attr == "SubsystemBase":
            import warnings

            warnings.warn(
                "SubsystemBase is deprecated", DeprecationWarning, stacklevel=2
            )
            return Subsystem

        if attr == "CommandBase":
            import warnings

            warnings.warn("CommandBase is deprecated", DeprecationWarning, stacklevel=2)
            return Command

        raise AttributeError(f"module {__name__!r} has no attribute {attr!r}")



# validated: 2024-01-20 DS 192a28af4731 SwerveControllerCommand.java
# Copyright (c) FIRST and other WPILib contributors.
# Open Source Software; you can modify and/or share it under the terms of
# the WPILib BSD license file in the root directory of this project.
from __future__ import annotations
from typing import Callable, Optional, Union, Tuple, Sequence

from wpimath.controller import (
    HolonomicDriveController,
)
from wpimath.geometry import Pose2d, Rotation2d
from wpimath.kinematics import (
    SwerveDrive2Kinematics,
    SwerveDrive3Kinematics,
    SwerveDrive4Kinematics,
    SwerveDrive6Kinematics,
    SwerveModuleState,
)
from wpimath.trajectory import Trajectory
from wpilib import Timer

from .command import Command
from .subsystem import Subsystem


class SwerveControllerCommand(Command):
    """
    A command that uses two PID controllers (:class:`wpimath.controller.PIDController`)
    and a HolonomicDriveController (:class:`wpimath.controller.HolonomicDriveController`)
    to follow a trajectory (:class:`wpimath.trajectory.Trajectory`) with a swerve drive.

    This command outputs the raw desired Swerve Module States (:class:`wpimath.kinematics.SwerveModuleState`) in an
    array. The desired wheel and module rotation velocities should be taken from those and used in
    velocity PIDs.

    The robot angle controller does not follow the angle given by the trajectory but rather goes
    to the angle given in the final state of the trajectory.
    """

    def __init__(
        self,
        trajectory: Trajectory,
        pose: Callable[[], Pose2d],
        kinematics: Union[
            SwerveDrive2Kinematics,
            SwerveDrive3Kinematics,
            SwerveDrive4Kinematics,
            SwerveDrive6Kinematics,
        ],
        controller: HolonomicDriveController,
        outputModuleStates: Callable[[Sequence[SwerveModuleState]], None],
        requirements: Tuple[Subsystem],
        desiredRotation: Optional[Callable[[], Rotation2d]] = None,
    ) -> None:
        """
        Constructs a new SwerveControllerCommand that when executed will follow the
        provided trajectory. This command will not return output voltages but
        rather raw module states from the position controllers which need to be put
        into a velocity PID.

        Note: The controllers will *not* set the outputVolts to zero upon
        completion of the path- this is left to the user, since it is not
        appropriate for paths with nonstationary endstates.

        :param trajectory:         The trajectory to follow.
        :param pose:               A function that supplies the robot pose - use one of the odometry classes to
                                   provide this.
        :param kinematics:         The kinematics for the robot drivetrain. Can be kinematics for 2/3/4/6
                                   SwerveKinematics.
        :param controller:         The HolonomicDriveController for the drivetrain.
                                   If you have x, y, and theta controllers, pass them into
                                   HolonomicPIDController.
        :param outputModuleStates: The raw output module states from the position controllers.
        :param requirements:       The subsystems to require.
        :param desiredRotation:    (optional) The angle that the drivetrain should be
                                   facing. This is sampled at each time step. If not specified, that rotation of
                                   the final pose in the trajectory is used.
        """
        super().__init__()
        self._trajectory = trajectory
        self._pose = pose
        self._kinematics = kinematics
        self._outputModuleStates = outputModuleStates
        self._controller = controller
        if desiredRotation is None:
            self._desiredRotation = trajectory.states()[-1].pose.rotation
        else:
            self._desiredRotation = desiredRotation

        self._timer = Timer()
        self.addRequirements(*requirements)

    def initialize(self):
        self._timer.restart()

    def execute(self):
        curTime = self._timer.get()
        desiredState = self._trajectory.sample(curTime)

        targetChassisSpeeds = self._controller.calculate(
            self._pose(), desiredState, self._desiredRotation()
        )
        targetModuleStates = self._kinematics.toSwerveModuleStates(targetChassisSpeeds)

        self._outputModuleStates(targetModuleStates)

    def end(self, interrupted):
        self._timer.stop()

    def isFinished(self):
        return self._timer.hasElapsed(self._trajectory.totalTime())



from typing import TYPE_CHECKING

from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *

import commands2
import pytest


def test_timeout(scheduler: commands2.CommandScheduler):
    with ManualSimTime() as sim:
        command1 = commands2.WaitCommand(1)
        timeout = command1.withTimeout(2)
        scheduler.schedule(timeout)
        scheduler.run()
        assert not command1.isScheduled()
        assert timeout.isScheduled()
        sim.step(3)
        scheduler.run()
        assert not timeout.isScheduled()


def test_until(scheduler: commands2.CommandScheduler):
    condition = OOBoolean(False)
    command = commands2.WaitCommand(10).until(condition)
    scheduler.schedule(command)
    scheduler.run()
    assert command.isScheduled()
    condition.set(True)
    scheduler.run()
    assert not command.isScheduled()


def test_only_while(scheduler: commands2.CommandScheduler):
    condition = OOBoolean(True)
    command = commands2.WaitCommand(10).onlyWhile(condition)
    scheduler.schedule(command)
    scheduler.run()
    assert command.isScheduled()
    condition.set(False)
    scheduler.run()
    assert not command.isScheduled()


def test_ignoringDisable(scheduler: commands2.CommandScheduler):
    command = commands2.RunCommand(lambda: None).ignoringDisable(True)
    DriverStationSim.setEnabled(False)
    DriverStationSim.notifyNewData()
    scheduler.schedule(command)
    scheduler.run()
    assert command.isScheduled()


def test_beforeStarting(scheduler: commands2.CommandScheduler):
    condition = OOBoolean(False)
    condition.set(False)
    command = commands2.InstantCommand()
    scheduler.schedule(
        command.beforeStarting(commands2.InstantCommand(lambda: condition.set(True)))
    )
    assert condition == True


@pytest.mark.skip
def test_andThenLambda(scheduler: commands2.CommandScheduler):
    ...


def test_andThen(scheduler: commands2.CommandScheduler):
    condition = OOBoolean(False)
    condition.set(False)
    command1 = commands2.InstantCommand()
    command2 = commands2.InstantCommand(lambda: condition.set(True))
    scheduler.schedule(command1.andThen(command2))
    assert condition == False
    scheduler.run()
    assert condition == True


def test_deadlineWith(scheduler: commands2.CommandScheduler):
    condition = OOBoolean(False)
    condition.set(False)

    dictator = commands2.WaitUntilCommand(condition)
    endsBefore = commands2.InstantCommand()
    endsAfter = commands2.WaitUntilCommand(lambda: False)

    group = dictator.deadlineWith(endsBefore, endsAfter)

    scheduler.schedule(group)
    scheduler.run()
    assert group.isScheduled()
    condition.set(True)
    scheduler.run()
    assert not group.isScheduled()


def test_alongWith(scheduler: commands2.CommandScheduler):
    condition = OOBoolean()
    condition.set(False)

    command1 = commands2.WaitUntilCommand(condition)
    command2 = commands2.InstantCommand()

    group = command1.alongWith(command2)

    scheduler.schedule(group)
    scheduler.run()
    assert group.isScheduled()
    condition.set(True)
    scheduler.run()
    assert not group.isScheduled()


def test_raceWith(scheduler: commands2.CommandScheduler):
    command1 = commands2.WaitUntilCommand(lambda: False)
    command2 = commands2.InstantCommand()

    group = command1.raceWith(command2)

    scheduler.schedule(group)
    scheduler.run()
    assert not group.isScheduled()


def test_unless(scheduler: commands2.CommandScheduler):
    unlessCondition = OOBoolean(True)
    hasRunCondition = OOBoolean(False)

    command = commands2.InstantCommand(lambda: hasRunCondition.set(True)).unless(
        unlessCondition
    )

    scheduler.schedule(command)
    scheduler.run()
    assert hasRunCondition == False
    unlessCondition.set(False)
    scheduler.schedule(command)
    scheduler.run()
    assert hasRunCondition == True


def test_onlyIf(scheduler: commands2.CommandScheduler):
    onlyIfCondition = OOBoolean(False)
    hasRunCondition = OOBoolean(False)

    command = commands2.InstantCommand(lambda: hasRunCondition.set(True)).onlyIf(
        onlyIfCondition
    )

    scheduler.schedule(command)
    scheduler.run()
    assert hasRunCondition == False
    onlyIfCondition.set(True)
    scheduler.schedule(command)
    scheduler.run()
    assert hasRunCondition == True


def test_finallyDo(scheduler: commands2.CommandScheduler):
    first = OOInteger(0)
    second = OOInteger(0)

    command = commands2.FunctionalCommand(
        lambda: None,
        lambda: None,
        lambda interrupted: first.incrementAndGet() if not interrupted else None,
        lambda: True,
    ).finallyDo(lambda interrupted: second.addAndGet(1 + first()))

    scheduler.schedule(command)
    assert first == 0
    assert second == 0
    scheduler.run()
    assert first == 1
    assert second == 2


def test_handleInterrupt(scheduler: commands2.CommandScheduler):
    first = OOInteger(0)
    second = OOInteger(0)

    command = commands2.FunctionalCommand(
        lambda: None,
        lambda: None,
        lambda interrupted: first.incrementAndGet() if interrupted else None,
        lambda: False,
    ).handleInterrupt(lambda: second.addAndGet(1 + first()))

    scheduler.schedule(command)
    scheduler.run()
    assert first == 0
    assert second == 0
    scheduler.cancel(command)
    assert first == 1
    assert second == 2


def test_withName(scheduler: commands2.CommandScheduler):
    command = commands2.InstantCommand()
    name = "Named"
    named = command.withName(name)
    assert named.getName() == name



# validated: 2024-01-19 DS aaea85ff1656 ParallelCommandGroup.java
from __future__ import annotations

from typing import Dict

from .command import Command, InterruptionBehavior
from .commandscheduler import CommandScheduler
from .exceptions import IllegalCommandUse
from .util import flatten_args_commands


class ParallelCommandGroup(Command):
    """
    A command composition that runs a set of commands in parallel, ending when the last command ends.

    The rules for command compositions apply: command instances that are passed to it cannot be
    added to any other composition or scheduled individually, and the composition requires all
    subsystems its components require.
    """

    def __init__(self, *commands: Command):
        """
        Creates a new ParallelCommandGroup. The given commands will be executed simultaneously. The
        command composition will finish when the last command finishes. If the composition is
        interrupted, only the commands that are still running will be interrupted.

        :param commands: the commands to include in this composition.
        """
        super().__init__()
        self._commands: Dict[Command, bool] = {}
        self._runsWhenDisabled = True
        self._interruptBehavior = InterruptionBehavior.kCancelIncoming
        self.addCommands(*commands)

    def addCommands(self, *commands: Command):
        """
        Adds the given commands to the group.

        :param commands: Commands to add to the group
        """
        commands = flatten_args_commands(commands)
        if True in self._commands.values():
            raise IllegalCommandUse(
                "Commands cannot be added to a composition while it is running"
            )

        CommandScheduler.getInstance().registerComposedCommands(commands)

        for command in commands:
            in_common = command.getRequirements().intersection(self.requirements)
            if in_common:
                raise IllegalCommandUse(
                    "Multiple commands in a parallel composition cannot require the same subsystems.",
                    common=in_common,
                )

            self._commands[command] = False
            self.requirements.update(command.getRequirements())
            self._runsWhenDisabled = (
                self._runsWhenDisabled and command.runsWhenDisabled()
            )

            if command.getInterruptionBehavior() == InterruptionBehavior.kCancelSelf:
                self._interruptBehavior = InterruptionBehavior.kCancelSelf

    def initialize(self):
        for command in self._commands:
            command.initialize()
            self._commands[command] = True

    def execute(self):
        for command, isRunning in self._commands.items():
            if not isRunning:
                continue
            command.execute()
            if command.isFinished():
                command.end(False)
                self._commands[command] = False

    def end(self, interrupted: bool):
        if interrupted:
            for command, isRunning in self._commands.items():
                if not isRunning:
                    continue
                command.end(True)
                self._commands[command] = False

    def isFinished(self) -> bool:
        return True not in self._commands.values()

    def runsWhenDisabled(self) -> bool:
        return self._runsWhenDisabled

    def getInterruptionBehavior(self) -> InterruptionBehavior:
        return self._interruptBehavior



from typing import TYPE_CHECKING

import commands2
from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *

import pytest


def test_runCommandSchedule(scheduler: commands2.CommandScheduler):
    counter = OOInteger(0)

    command = commands2.RunCommand(counter.incrementAndGet)

    scheduler.schedule(command)
    scheduler.run()
    scheduler.run()
    scheduler.run()

    assert counter == 3



from typing import Any, Dict, TypeVar, Type

import inspect

import commands2
from wpilib.simulation import DriverStationSim, pauseTiming, resumeTiming, stepTiming


Y = TypeVar("Y")


def full_subclass_of(cls: Type[Y]) -> Type[Y]:
    # Pybind classes can't be monkeypatched.
    # This generates a subclass with every method filled out
    # so that it can be monkeypatched.
    retlist = []
    clsname = cls.__name__  # + "_Subclass"
    classdef = f"class {clsname}(cls):\n"
    for name in dir(cls):
        # for name in set(dir(cls)):
        value = getattr(cls, name)
        if callable(value) and not inspect.isclass(value) and not name.startswith("_"):
            classdef += f"  def {name}(self, *args, **kwargs):\n"
            classdef += f"    return super().{name}(*args, **kwargs)\n"
    classdef += "  ...\n"
    classdef += f"retlist.append({clsname})\n"
    print(classdef)
    exec(classdef, globals(), locals())
    return retlist[0]


class ManualSimTime:
    def __enter__(self) -> "ManualSimTime":
        pauseTiming()
        return self

    def __exit__(self, *args):
        resumeTiming()

    def step(self, delta: float):
        stepTiming(delta)


class OOInteger:
    def __init__(self, value: int = 0) -> None:
        self.value = value

    def get(self) -> int:
        return self.value

    def set(self, value: int):
        self.value = value

    def incrementAndGet(self) -> int:
        self.value += 1
        return self.value

    def addAndGet(self, value: int) -> int:
        self.value += value
        return self.value

    def __eq__(self, value: float) -> bool:
        return self.value == value

    def __lt__(self, value: float) -> bool:
        return self.value < value

    def __call__(self) -> int:
        return self.value


class OOBoolean:
    def __init__(self, value: bool = False) -> None:
        self.value = value

    def get(self) -> bool:
        return self.value

    def set(self, value: bool):
        self.value = value

    def __eq__(self, value: object) -> bool:
        return self.value == value

    def __bool__(self) -> bool:
        return self.value

    def __call__(self) -> bool:
        return self.value


class InternalButton(commands2.button.Trigger):
    def __init__(self):
        super().__init__(self.isPressed)
        self.pressed = False

    def isPressed(self) -> bool:
        return self.pressed

    def setPressed(self, value: bool):
        self.pressed = value

    def __call__(self) -> bool:
        return self.pressed


class OOFloat:
    def __init__(self, value: float = 0.0) -> None:
        self.value = value

    def get(self) -> float:
        return self.value

    def set(self, value: float):
        self.value = value

    def incrementAndGet(self) -> float:
        self.value += 1
        return self.value

    def addAndGet(self, value: float) -> float:
        self.value += value
        return self.value

    def __eq__(self, value: float) -> bool:
        return self.value == value

    def __lt__(self, value: float) -> bool:
        return self.value < value

    def __call__(self) -> float:
        return self.value

    def __name__(self) -> str:
        return "OOFloat"


##########################################
# Fakito Framework


def _get_all_args_as_kwargs(method, *args, **kwargs) -> Dict[str, Any]:
    try:
        import inspect

        method_args = inspect.getcallargs(method, *args, **kwargs)

        method_arg_names = list(inspect.signature(method).parameters.keys())

        for idx, arg in enumerate(args):
            method_args[method_arg_names[idx]] = arg

        try:
            del method_args["self"]
        except KeyError:
            pass
        return method_args
    except TypeError:
        # Pybind methods can't be inspected
        # The exact args/kwargs that are passed in are checked instead
        r = {}
        for idx, arg in enumerate(args):
            r[idx] = arg
        r.update(kwargs)
        return r


class MethodWrapper:
    def __init__(self, method):
        self.method = method
        self.og_method = method
        self.times_called = 0
        self.call_log = []

    def __call__(self, *args, **kwargs):
        self.times_called += 1
        method_args = _get_all_args_as_kwargs(self.method, *args, **kwargs)
        self.call_log.append(method_args)
        return self.method(*args, **kwargs)

    def called_with(self, *args, **kwargs):
        return _get_all_args_as_kwargs(self.method, *args, **kwargs) in self.call_log

    def times_called_with(self, *args, **kwargs):
        return self.call_log.count(
            _get_all_args_as_kwargs(self.method, *args, **kwargs)
        )


def start_spying_on(obj: Any) -> None:
    """
    Mocks all methods on an object, so that that call info can be used in asserts.

    Example:
    ```
    obj = SomeClass()
    start_spying_on(obj)
    obj.method()
    obj.method = lambda: None # supports monkeypatching
    assert obj.method.times_called == 2
    assert obj.method.called_with(arg1=1, arg2=2)
    assert obj.method.times_called_with(arg1=1, arg2=2) == 2
    ```
    """

    for name in dir(obj):
        value = getattr(obj, name)
        if callable(value) and not inspect.isclass(value) and not name.startswith("_"):
            setattr(obj, name, MethodWrapper(value))

    if not hasattr(obj.__class__, "_is_being_spied_on"):
        try:
            old_setattr = obj.__class__.__setattr__
        except AttributeError:
            old_setattr = object.__setattr__

        def _setattr(self, name, value):
            if name in dir(self):
                existing_value = getattr(self, name)
                if isinstance(existing_value, MethodWrapper):
                    existing_value.method = value
                    return
            old_setattr(self, name, value)

        obj.__class__.__setattr__ = _setattr
        obj.__class__._is_being_spied_on = True


# fakito verify
def reset(obj: Any) -> None:
    """
    Resets the call log of all mocked methods on an object.
    Also restores all monkeypatched methods.
    """
    for name in dir(obj):
        value = getattr(obj, name)
        if isinstance(value, MethodWrapper):
            value.method = value.og_method
            value.times_called = 0
            value.call_log = []


class times:
    def __init__(self, times: int) -> None:
        self.times = times


def never() -> times:
    return times(0)


class _verify:
    def __init__(self, obj: Any, times: times = times(1)):
        self.obj = obj
        self.times = times.times

    def __getattribute__(self, name: str) -> Any:
        def self_dot(name: str):
            return super(_verify, self).__getattribute__(name)

        def times_string(times: int) -> str:
            if times == 1:
                return "1 time"
            else:
                return f"{times} times"

        def check(*args, **kwargs):
            __tracebackhide__ = True
            # import code
            # code.interact(local={**globals(), **locals()})
            method = getattr(self_dot("obj"), name)
            # method = getattr(self1.obj, name)
            assert method.times_called_with(*args, **kwargs) == self_dot(
                "times"
            ), f"Expected {name} to be called {times_string(self_dot('times'))} with {args} {kwargs}, but was called {times_string(method.times_called_with(*args, **kwargs))}"

        return check


T = TypeVar("T")


def verify(obj: T, times: times = times(1)) -> T:
    # import code
    # code.interact(local={**globals(), **locals()})
    return _verify(obj, times)  # type: ignore


def patch_via_decorator(obj: Any):
    def decorator(method):
        setattr(obj, method.__name__, method.__get__(obj, obj.__class__))
        return method

    return decorator



# validated: 2024-01-24 DS 192a28af4731 TrapezoidProfileCommand.java
#
# Copyright (c) FIRST and other WPILib contributors.
# Open Source Software; you can modify and/or share it under the terms of
# the WPILib BSD license file in the root directory of this project.
#
from __future__ import annotations

from typing import Callable, Any

from wpilib import Timer

from .command import Command
from .subsystem import Subsystem


class TrapezoidProfileCommand(Command):
    """
    A command that runs a :class:`wpimath.trajectory.TrapezoidProfile`. Useful for smoothly controlling mechanism motion.
    """

    def __init__(
        self,
        profile,
        output: Callable[[Any], Any],
        goal: Callable[[], Any],
        currentState: Callable[[], Any],
        *requirements: Subsystem,
    ):
        """Creates a new TrapezoidProfileCommand that will execute the given :class:`wpimath.trajectory.TrapezoidProfile`.
        Output will be piped to the provided consumer function.

        :param profile:      The motion profile to execute.
        :param output:       The consumer for the profile output.
        :param goal:         The supplier for the desired state
        :param currentState: The supplier for the current state
        :param requirements: The subsystems required by this command.
        """
        super().__init__()
        self._profile = profile
        self._output = output
        self._goal = goal
        self._currentState = currentState
        self._timer = Timer()

        self.addRequirements(*requirements)

    def initialize(self) -> None:
        self._timer.restart()

    def execute(self) -> None:
        self._output(
            self._profile.calculate(
                self._timer.get(),
                self._currentState(),
                self._goal(),
            )
        )

    def end(self, interrupted) -> None:
        self._timer.stop()

    def isFinished(self) -> bool:
        return self._timer.hasElapsed(self._profile.totalTime())



#!/usr/bin/env python3

import setuptools

setuptools.setup(
    name="robotpy-commands-v2",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    author="RobotPy Development Team",
    author_email="robotpy@googlegroups.com",
    description="WPILib command framework v2",
    url="https://github.com/robotpy/robotpy-commands-v2",
    packages=["commands2"],
    install_requires=[
        "wpilib<2025,>=2024.2.1.2",
        "typing_extensions>=4.1.0,<5",
    ],
    license="BSD-3-Clause",
    python_requires=">=3.8",
    include_package_data=True,
)



# notrack
from wpilib import TimedRobot

from .commandscheduler import CommandScheduler

seconds = float


class TimedCommandRobot(TimedRobot):
    kSchedulerOffset = 0.005

    def __init__(self, period: seconds = TimedRobot.kDefaultPeriod / 1000) -> None:
        super().__init__(period)
        self.addPeriodic(
            CommandScheduler.getInstance().run, period, self.kSchedulerOffset
        )



# validated: 2024-01-20 DS f29a7d2e501b Command.java
# Don't import stuff from the package here, it'll cause a circular import


from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Set, Union

from typing_extensions import Self, TypeAlias

if TYPE_CHECKING:
    from .instantcommand import InstantCommand
    from .subsystem import Subsystem
    from .parallelracegroup import ParallelRaceGroup
    from .sequentialcommandgroup import SequentialCommandGroup
    from .paralleldeadlinegroup import ParallelDeadlineGroup
    from .parallelcommandgroup import ParallelCommandGroup
    from .repeatcommand import RepeatCommand
    from .proxycommand import ProxyCommand
    from .conditionalcommand import ConditionalCommand
    from .wrappercommand import WrapperCommand

from wpiutil import Sendable, SendableRegistry, SendableBuilder


class InterruptionBehavior(Enum):
    """
    An enum describing the command's behavior when another command with a shared requirement is
    scheduled.
    """

    kCancelIncoming = 0
    """
    This command ends, :func:`commands2.Command.end` is called, and the incoming command is
    scheduled normally.

    This is the default behavior.
    """

    kCancelSelf = 1
    """ This command continues, and the incoming command is not scheduled."""


class Command(Sendable):
    """
    A state machine representing a complete action to be performed by the robot. Commands are run by
    the :class:`commands2.CommandScheduler`, and can be composed into CommandGroups to allow users to build
    complicated multistep actions without the need to roll the state machine logic themselves.

    Commands are run synchronously from the main robot loop; no multithreading is used, unless
    specified explicitly from the command implementation.
    """

    InterruptionBehavior: TypeAlias = (
        InterruptionBehavior  # type alias for 2023 location
    )

    requirements: Set[Subsystem]

    def __new__(cls, *args, **kwargs) -> Self:
        instance = super().__new__(
            cls,
        )
        super().__init__(instance)
        SendableRegistry.add(instance, cls.__name__)
        instance.requirements = set()
        return instance

    def __init__(self):
        pass

    def initialize(self):
        """The initial subroutine of a command. Called once when the command is initially scheduled."""
        pass

    def execute(self):
        """The main body of a command. Called repeatedly while the command is scheduled."""
        pass

    def end(self, interrupted: bool):
        """
        The action to take when the command ends. Called when either the command finishes normally, or
        when it interrupted/canceled.

        Do not schedule commands here that share requirements with this command. Use :meth:`.andThen` instead.

        :param interrupted: whether the command was interrupted/canceled
        """
        pass

    def isFinished(self) -> bool:
        """
        Whether the command has finished. Once a command finishes, the scheduler will call its :meth:`commands2.Command.end`
        method and un-schedule it.

        :returns: whether the command has finished.
        """
        return False

    def getRequirements(self) -> Set[Subsystem]:
        """
        Specifies the set of subsystems used by this command. Two commands cannot use the same
        subsystem at the same time. If another command is scheduled that shares a requirement, :meth:`.getInterruptionBehavior` will be checked and followed. If no subsystems are required, return
        an empty set.

        .. note:: it is recommended that user implementations contain the requirements as a field, and
                  return that field here, rather than allocating a new set every time this is called.

        :returns: the set of subsystems that are required
        """
        return self.requirements

    def addRequirements(self, *requirements: Subsystem):
        """
        Adds the specified subsystems to the requirements of the command. The scheduler will prevent
        two commands that require the same subsystem from being scheduled simultaneously.

        .. note:: The scheduler determines the requirements of a command when it is scheduled, so
                  this method should normally be called from the command's constructor.

        :param requirements: the requirements to add
        """
        self.requirements.update(requirements)

    def getName(self) -> str:
        """
        Gets the name of this Command.

        :returns: Name
        """
        return SendableRegistry.getName(self)

    def setName(self, name: str):
        """
        Sets the name of this Command.

        :param name: Name
        """
        SendableRegistry.setName(self, name)

    def getSubsystem(self) -> str:
        """
        Gets the subsystem name of this Subsystem.

        :returns: Subsystem name
        """
        return SendableRegistry.getSubsystem(self)

    def setSubsystem(self, subsystem: str):
        """
        Sets the subsystem name of this Subsystem.

        :param subsystem: subsystem name
        """
        SendableRegistry.setSubsystem(self, subsystem)

    def withTimeout(self, seconds: float) -> ParallelRaceGroup:
        """
        Decorates this command with a timeout. If the specified timeout is exceeded before the command
        finishes normally, the command will be interrupted and un-scheduled. Note that the timeout only
        applies to the command returned by this method; the calling command is not itself changed.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param seconds: the timeout duration
        :returns: the command with the timeout added
        """
        from .waitcommand import WaitCommand

        return self.raceWith(WaitCommand(seconds))

    def until(self, condition: Callable[[], bool]) -> ParallelRaceGroup:
        """
        Decorates this command with an interrupt condition. If the specified condition becomes true
        before the command finishes normally, the command will be interrupted and un-scheduled.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param condition: the interrupt condition
        :returns: the command with the interrupt condition added
        """
        from .waituntilcommand import WaitUntilCommand

        return self.raceWith(WaitUntilCommand(condition))

    def onlyWhile(self, condition: Callable[[], bool]) -> ParallelRaceGroup:
        """
        Decorates this command with a run condition. If the specified condition becomes false before
        the command finishes normally, the command will be interrupted and un-scheduled.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param condition: the interrupt condition
        :returns: the command with the interrupt condition added
        """
        assert callable(condition)
        return self.until(lambda: not condition())

    def beforeStarting(
        self, before: Union[Command, Callable[[], None]]
    ) -> SequentialCommandGroup:
        """
        Decorates this command with a callable or command to run before this command starts.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param before: the command to run before this one
        :returns: the decorated command
        """
        from .instantcommand import InstantCommand
        from .sequentialcommandgroup import SequentialCommandGroup

        if callable(before):
            before = InstantCommand(before)

        return SequentialCommandGroup(before, self)

    def andThen(self, *next: Command) -> SequentialCommandGroup:
        """
        Decorates this command with a set of commands to run after it in sequence. Often more
        convenient/less-verbose than constructing a new SequentialCommandGroup explicitly.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param next: the commands to run next
        :returns: the decorated command
        """
        from .sequentialcommandgroup import SequentialCommandGroup

        return SequentialCommandGroup(self, *next)

    def deadlineWith(self, *parallel: Command) -> ParallelDeadlineGroup:
        """
        Decorates this command with a set of commands to run parallel to it, ending when the calling
        command ends and interrupting all the others. Often more convenient/less-verbose than
        constructing a new ParallelDeadlineGroup explicitly.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param parallel: the commands to run in parallel
        :returns: the decorated command
        """
        from .paralleldeadlinegroup import ParallelDeadlineGroup

        return ParallelDeadlineGroup(self, *parallel)

    def alongWith(self, *parallel: Command) -> ParallelCommandGroup:
        """
        Decorates this command with a set of commands to run parallel to it, ending when the last
        command ends. Often more convenient/less-verbose than constructing a new {@link
        ParallelCommandGroup} explicitly.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param parallel: the commands to run in parallel
        :returns: the decorated command
        """
        from .parallelcommandgroup import ParallelCommandGroup

        return ParallelCommandGroup(self, *parallel)

    def raceWith(self, *parallel: Command) -> ParallelRaceGroup:
        """
        Decorates this command with a set of commands to run parallel to it, ending when the first
        command ends. Often more convenient/less-verbose than constructing a new {@link
        ParallelRaceGroup} explicitly.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param parallel: the commands to run in parallel
        :returns: the decorated command
        """
        from .parallelracegroup import ParallelRaceGroup

        return ParallelRaceGroup(self, *parallel)

    def repeatedly(self) -> RepeatCommand:
        """
        Decorates this command to run repeatedly, restarting it when it ends, until this command is
        interrupted. The decorated command can still be canceled.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :returns: the decorated command
        """
        from .repeatcommand import RepeatCommand

        return RepeatCommand(self)

    def asProxy(self) -> ProxyCommand:
        """
        Decorates this command to run "by proxy" by wrapping it in a ProxyCommand. This is
        useful for "forking off" from command compositions when the user does not wish to extend the
        command's requirements to the entire command composition.

        :returns: the decorated command
        """
        from .proxycommand import ProxyCommand

        return ProxyCommand(self)

    def unless(self, condition: Callable[[], bool]) -> ConditionalCommand:
        """
        Decorates this command to only run if this condition is not met. If the command is already
        running and the condition changes to true, the command will not stop running. The requirements
        of this command will be kept for the new conditional command.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param condition: the condition that will prevent the command from running
        :returns: the decorated command
        """
        from .conditionalcommand import ConditionalCommand
        from .instantcommand import InstantCommand

        return ConditionalCommand(InstantCommand(), self, condition)

    def onlyIf(self, condition: Callable[[], bool]) -> ConditionalCommand:
        """
        Decorates this command to only run if this condition is met. If the command is already running
        and the condition changes to false, the command will not stop running. The requirements of this
        command will be kept for the new conditional command.

        .. note:: This decorator works by adding this command to a composition.
                  The command the decorator was called on cannot be scheduled
                  independently or be added to a different composition (namely,
                  decorators), unless it is manually cleared from the list of composed
                  commands with :func:`commands2.CommandScheduler.removeComposedCommand`.
                  The command composition returned from this method can be further
                  decorated without issue.

        :param condition: the condition that will allow the command to run
        :returns: the decorated command
        """
        return self.unless(lambda: not condition())

    def ignoringDisable(self, doesRunWhenDisabled: bool) -> WrapperCommand:
        """
        Decorates this command to run or stop when disabled.

        :param doesRunWhenDisabled: true to run when disabled.
        :returns: the decorated command
        """
        from .wrappercommand import WrapperCommand

        class W(WrapperCommand):
            def runsWhenDisabled(self) -> bool:
                return doesRunWhenDisabled

        return W(self)

    def withInterruptBehavior(self, behavior: InterruptionBehavior) -> WrapperCommand:
        """
        Decorates this command to have a different InterruptionBehavior interruption behavior.

        :param interruptBehavior: the desired interrupt behavior
        :returns: the decorated command
        """
        from .wrappercommand import WrapperCommand

        class W(WrapperCommand):
            def getInterruptionBehavior(self) -> InterruptionBehavior:
                return behavior

        return W(self)

    def finallyDo(self, end: Callable[[bool], Any]) -> WrapperCommand:
        """
        Decorates this command with a lambda to call on interrupt or end, following the command's
        inherent :func:`commands2.Command.end` method.

        :param end: a lambda accepting a boolean parameter specifying whether the command was
            interrupted.
        :return: the decorated command
        """
        from .wrappercommand import WrapperCommand

        class W(WrapperCommand):
            def end(self, interrupted: bool) -> None:
                self._command.end(interrupted)
                end(interrupted)

        return W(self)

    def handleInterrupt(self, handler: Callable[[], Any]) -> WrapperCommand:
        """
        Decorates this command with a lambda to call on interrupt, following the command's inherent
        :func:`commands2.Command.end` method.

        :param handler: a lambda to run when the command is interrupted
        :returns: the decorated command
        """
        return self.finallyDo(lambda interrupted: handler() if interrupted else None)

    def schedule(self) -> None:
        """Schedules this command."""
        from .commandscheduler import CommandScheduler

        CommandScheduler.getInstance().schedule(self)

    def cancel(self) -> None:
        """
        Cancels this command. Will call ``end(true)``. Commands will be canceled
        regardless of InterruptionBehavior interruption behavior.
        """
        from .commandscheduler import CommandScheduler

        CommandScheduler.getInstance().cancel(self)

    def isScheduled(self) -> bool:
        """
        Whether the command is currently scheduled. Note that this does not detect whether the command
        is in a composition, only whether it is directly being run by the scheduler.

        :returns: Whether the command is scheduled.
        """
        from .commandscheduler import CommandScheduler

        return CommandScheduler.getInstance().isScheduled(self)

    def hasRequirement(self, requirement: Subsystem) -> bool:
        """
        Whether the command requires a given subsystem.

        :param requirement: the subsystem to inquire about
        :returns: whether the subsystem is required
        """
        return requirement in self.requirements

    def getInterruptionBehavior(self) -> InterruptionBehavior:
        """
        How the command behaves when another command with a shared requirement is scheduled.

        :returns: a variant of InterruptionBehavior, defaulting to {@link InterruptionBehavior#kCancelSelf kCancelSelf}.
        """
        return InterruptionBehavior.kCancelSelf

    def runsWhenDisabled(self) -> bool:
        """
        Whether the given command should run when the robot is disabled. Override to return true if the
        command should run when disabled.

        :returns: whether the command should run when the robot is disabled
        """
        return False

    def withName(self, name: str) -> WrapperCommand:
        """
        Decorates this command with a name.

        :param name: the name of the command
        :returns: the decorated command
        """
        from .wrappercommand import WrapperCommand

        wrapper = WrapperCommand(self)
        wrapper.setName(name)
        return wrapper

    def initSendable(self, builder: SendableBuilder) -> None:
        from .commandscheduler import CommandScheduler

        builder.setSmartDashboardType("Command")
        builder.addStringProperty(
            ".name",
            self.getName,
            lambda _: None,
        )

        def on_set(value: bool) -> None:
            if value:
                if not self.isScheduled():
                    self.schedule()
            else:
                if self.isScheduled():
                    self.cancel()

        builder.addBooleanProperty(
            "running",
            self.isScheduled,
            on_set,
        )
        builder.addBooleanProperty(
            ".isParented",
            lambda: CommandScheduler.getInstance().isComposed(self),
            lambda _: None,
        )
        builder.addStringProperty(
            "interruptBehavior",
            lambda: self.getInterruptionBehavior().name,
            lambda _: None,
        )
        builder.addBooleanProperty(
            "runsWhenDisabled",
            self.runsWhenDisabled,
            lambda _: None,
        )



from typing import TYPE_CHECKING

import commands2
from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *

import pytest


def test_instantCommandSchedule(scheduler: commands2.CommandScheduler):
    cond = OOBoolean()

    command = commands2.InstantCommand(lambda: cond.set(True))

    scheduler.schedule(command)
    scheduler.run()

    assert cond
    assert not scheduler.isScheduled(command)



# validated: 2024-01-20 DS 1144115da01f Subsystem.java
# Don't import stuff from the package here, it'll cause a circular import

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional
from typing_extensions import Self

if TYPE_CHECKING:
    from .command import Command
    from .commandscheduler import CommandScheduler

from wpiutil import Sendable, SendableBuilder, SendableRegistry


class Subsystem(Sendable):
    """
    A robot subsystem. Subsystems are the basic unit of robot organization in the Command-based
    framework; they encapsulate low-level hardware objects (motor controllers, sensors, etc.) and
    provide methods through which they can be used by Commands. Subsystems are used by the
    CommandScheduler's resource management system to ensure multiple robot actions are not
    "fighting" over the same hardware; Commands that use a subsystem should include that subsystem in
    their :func:`commands2.Command.getRequirements` method, and resources used within a subsystem should
    generally remain encapsulated and not be shared by other parts of the robot.

    Subsystems must be registered with the scheduler with the :func:`commands2.CommandScheduler.registerSubsystem`
    method in order for the :func:`.periodic` method to be called. It is recommended that this method be called from the
    constructor of users' Subsystem implementations.
    """

    def __new__(cls, *arg, **kwargs) -> Self:
        instance = super().__new__(cls)
        super().__init__(instance)
        SendableRegistry.addLW(instance, cls.__name__, cls.__name__)
        # add to the scheduler
        from .commandscheduler import CommandScheduler

        CommandScheduler.getInstance().registerSubsystem(instance)
        return instance

    def __init__(self) -> None:
        pass

    def periodic(self) -> None:
        """
        This method is called periodically by the CommandScheduler. Useful for updating
        subsystem-specific state that you don't want to offload to a Command. Teams should try
        to be consistent within their own codebases about which responsibilities will be handled by
        Commands, and which will be handled here.
        """
        pass

    def simulationPeriodic(self) -> None:
        """
        This method is called periodically by the CommandScheduler. Useful for updating
        subsystem-specific state that needs to be maintained for simulations, such as for updating simulation classes and setting simulated sensor readings.
        """
        pass

    def setDefaultCommand(self, command: Command) -> None:
        """
        Sets the default Command of the subsystem. The default command will be automatically
        scheduled when no other commands are scheduled that require the subsystem. Default commands
        should generally not end on their own, i.e. their :func:`commands2.Command.isFinished` method should
        always return false. Will automatically register this subsystem with the CommandScheduler.

        :param defaultCommand: the default command to associate with this subsystem
        """
        from .commandscheduler import CommandScheduler

        CommandScheduler.getInstance().setDefaultCommand(self, command)

    def removeDefaultCommand(self) -> None:
        """
        Removes the default command for the subsystem. This will not cancel the default command if it
        is currently running.
        """
        CommandScheduler.getInstance().removeDefaultCommand(self)

    def getDefaultCommand(self) -> Optional[Command]:
        """
        Gets the default command for this subsystem. Returns None if no default command is currently
        associated with the subsystem.

        :returns: the default command associated with this subsystem
        """
        from .commandscheduler import CommandScheduler

        return CommandScheduler.getInstance().getDefaultCommand(self)

    def getCurrentCommand(self) -> Optional[Command]:
        """
        Returns the command currently running on this subsystem. Returns None if no command is
        currently scheduled that requires this subsystem.

        :returns: the scheduled command currently requiring this subsystem
        """
        from .commandscheduler import CommandScheduler

        return CommandScheduler.getInstance().requiring(self)

    def register(self):
        """
        Registers this subsystem with the :class:`.CommandScheduler`, allowing its
        :func:`.periodic` method to be called when the scheduler runs.
        """
        from .commandscheduler import CommandScheduler

        return CommandScheduler.getInstance().registerSubsystem(self)

    def runOnce(self, action: Callable[[], None]) -> Command:
        """
        Constructs a command that runs an action once and finishes. Requires this subsystem.

        :param action: the action to run
        :return: the command
        """
        from .cmd import runOnce

        return runOnce(action, self)

    def run(self, action: Callable[[], None]) -> Command:
        """
        Constructs a command that runs an action every iteration until interrupted. Requires this
        subsystem.

        :param action: the action to run
        :returns: the command"""
        from .cmd import run

        return run(action, self)

    def startEnd(self, start: Callable[[], None], end: Callable[[], None]) -> Command:
        """
        Constructs a command that runs an action once and another action when the command is
        interrupted. Requires this subsystem.

        :param start: the action to run on start
        :param end: the action to run on interrupt
        :returns: the command
        """
        from .cmd import startEnd

        return startEnd(start, end, self)

    def runEnd(self, run: Callable[[], None], end: Callable[[], None]) -> Command:
        """
        Constructs a command that runs an action every iteration until interrupted, and then runs a
        second action. Requires this subsystem.

        :param run: the action to run every iteration
        :param end: the action to run on interrupt
        :returns: the command
        """
        from .cmd import runEnd

        return runEnd(run, end, self)

    #
    # From SubsystemBase
    #

    def getName(self) -> str:
        """
        Gets the name of this Subsystem.

        :returns: Name
        """
        return SendableRegistry.getName(self)

    def setName(self, name: str) -> None:
        """
        Set the name of this Subsystem.
        """
        SendableRegistry.setName(self, name)

    def getSubsystem(self) -> str:
        """
        Gets the subsystem name of this Subsystem.

        :returns: Subsystem name
        """
        return SendableRegistry.getSubsystem(self)

    def setSubsystem(self, subsystem: str):
        """
        Sets the subsystem name of this Subsystem.

        :param subsystem: subsystem name
        """
        SendableRegistry.setSubsystem(self, subsystem)

    def addChild(self, name: str, child: Sendable) -> None:
        """
        Associates a :class:`wpiutil.Sendable` with this Subsystem. Also update the child's name.

        :param name:  name to give child
        :param child: sendable
        """
        SendableRegistry.addLW(child, self.getSubsystem(), name)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("Subsystem")

        builder.addBooleanProperty(
            ".hasDefault", lambda: self.getDefaultCommand() is not None, lambda _: None
        )

        def get_default():
            command = self.getDefaultCommand()
            if command is not None:
                return command.getName()
            return "none"

        builder.addStringProperty(".default", get_default, lambda _: None)
        builder.addBooleanProperty(
            ".hasCommand", lambda: self.getCurrentCommand() is not None, lambda _: None
        )

        def get_current():
            command = self.getCurrentCommand()
            if command is not None:
                return command.getName()
            return "none"

        builder.addStringProperty(".command", get_current, lambda _: None)



# validated: 2024-01-19 DS aaea85ff1656 ConditionalCommand.java
from __future__ import annotations

from typing import Callable, Optional

from wpiutil import SendableBuilder

from .command import Command, InterruptionBehavior
from .commandscheduler import CommandScheduler


class ConditionalCommand(Command):
    """
    A command composition that runs one of two commands, depending on the value of the given
    condition when this command is initialized.

    The rules for command compositions apply: command instances that are passed to it cannot be
    added to any other composition or scheduled individually, and the composition requires all
    subsystems its components require.
    """

    selectedCommand: Optional[Command]

    def __init__(
        self, onTrue: Command, onFalse: Command, condition: Callable[[], bool]
    ):
        """
        Creates a new ConditionalCommand.

        :param onTrue: the command to run if the condition is true
        :param onFalse: the command to run if the condition is false
        :param condition: the condition to determine which command to run
        """
        super().__init__()

        CommandScheduler.getInstance().registerComposedCommands([onTrue, onFalse])

        self.onTrue = onTrue
        self.onFalse = onFalse
        self.condition = condition

        self.addRequirements(*onTrue.getRequirements())
        self.addRequirements(*onFalse.getRequirements())

    def initialize(self):
        if self.condition():
            self.selectedCommand = self.onTrue
        else:
            self.selectedCommand = self.onFalse

        self.selectedCommand.initialize()

    def execute(self):
        assert self.selectedCommand is not None
        self.selectedCommand.execute()

    def end(self, interrupted: bool):
        assert self.selectedCommand is not None
        self.selectedCommand.end(interrupted)

    def isFinished(self) -> bool:
        assert self.selectedCommand is not None
        return self.selectedCommand.isFinished()

    def runsWhenDisabled(self) -> bool:
        return self.onTrue.runsWhenDisabled() and self.onFalse.runsWhenDisabled()

    def getInterruptionBehavior(self) -> InterruptionBehavior:
        if (
            self.onTrue.getInterruptionBehavior() == InterruptionBehavior.kCancelSelf
            or self.onFalse.getInterruptionBehavior()
            == InterruptionBehavior.kCancelSelf
        ):
            return InterruptionBehavior.kCancelSelf
        else:
            return InterruptionBehavior.kCancelIncoming

    def initSendable(self, builder: SendableBuilder):
        super().initSendable(builder)
        builder.addStringProperty("onTrue", self.onTrue.getName, lambda _: None)
        builder.addStringProperty("onFalse", self.onFalse.getName, lambda _: None)

        def _selected():
            if self.selectedCommand is None:
                return "null"
            else:
                return self.selectedCommand.getName()

        builder.addStringProperty(
            "selected",
            _selected,
            lambda _: None,
        )



# validated: 2024-01-19 DS 6e58db398d63 RepeatCommand.java
from __future__ import annotations

from wpiutil import SendableBuilder

from .command import Command, InterruptionBehavior
from .commandscheduler import CommandScheduler


class RepeatCommand(Command):
    """
    A command that runs another command repeatedly, restarting it when it ends, until this command is
    interrupted. Command instances that are passed to it cannot be added to any other groups, or
    scheduled individually.

    The rules for command compositions apply: command instances that are passed to it cannot be
    added to any other composition or scheduled individually,and the composition requires all
    subsystems its components require.
    """

    def __init__(self, command: Command):
        """
        Creates a new RepeatCommand. Will run another command repeatedly, restarting it whenever it
        ends, until this command is interrupted.

        :param command: the command to run repeatedly
        """
        super().__init__()
        self._command = command
        CommandScheduler.getInstance().registerComposedCommands([command])
        self.requirements.update(command.getRequirements())
        self.setName(f"Repeat({command.getName()})")

    def initialize(self):
        self._ended = False
        self._command.initialize()

    def execute(self):
        if self._ended:
            self._ended = False
            self._command.initialize()

        self._command.execute()

        if self._command.isFinished():
            # restart command
            self._command.end(False)
            self._ended = True

    def isFinished(self) -> bool:
        return False

    def end(self, interrupted: bool):
        # Make sure we didn't already call end() (which would happen if the command finished in the
        # last call to our execute())
        if not self._ended:
            self._command.end(interrupted)
            self._ended = True

    def runsWhenDisabled(self) -> bool:
        return self._command.runsWhenDisabled()

    def getInterruptionBehavior(self) -> InterruptionBehavior:
        return self._command.getInterruptionBehavior()

    def initSendable(self, builder: SendableBuilder) -> None:
        super().initSendable(builder)
        builder.addStringProperty("command", self._command.getName, lambda _: None)



# validated: 2024-01-19 DS 6e58db398d63 FunctionalCommand.java
from __future__ import annotations

from typing import Any, Callable

from .command import Command
from .subsystem import Subsystem


class FunctionalCommand(Command):
    """
    A command that allows the user to pass in functions for each of the basic command methods through
    the constructor. Useful for inline definitions of complex commands - note, however, that if a
    command is beyond a certain complexity it is usually better practice to write a proper class for
    it than to inline it.
    """

    def __init__(
        self,
        onInit: Callable[[], Any],
        onExecute: Callable[[], Any],
        onEnd: Callable[[bool], Any],
        isFinished: Callable[[], bool],
        *requirements: Subsystem,
    ):
        """
        Creates a new FunctionalCommand.

        :param onInit: the function to run on command initialization
        :param onExecute: the function to run on command execution
        :param onEnd: the function to run on command end
        :param isFinished: the function that determines whether the command has finished
        :param requirements: the subsystems required by this command
        """
        super().__init__()

        assert callable(onInit)
        assert callable(onExecute)
        assert callable(onEnd)
        assert callable(isFinished)

        self._onInit = onInit
        self._onExecute = onExecute
        self._onEnd = onEnd
        self._isFinished = isFinished

        self.addRequirements(*requirements)

    def initialize(self):
        self._onInit()

    def execute(self):
        self._onExecute()

    def end(self, interrupted: bool):
        self._onEnd(interrupted)

    def isFinished(self) -> bool:
        return self._isFinished()



from typing import TYPE_CHECKING

import commands2
from compositiontestbase import MultiCompositionTestBase  # type: ignore
from util import *  # type: ignore

if TYPE_CHECKING:
    from .util import *
    from .compositiontestbase import MultiCompositionTestBase

import pytest


class TestSelectCommandComposition(MultiCompositionTestBase):
    def compose(self, *members: commands2.Command):
        return commands2.SelectCommand(dict(enumerate(members)), lambda: 0)


def test_selectCommand(scheduler: commands2.CommandScheduler):
    command1 = commands2.Command()
    command2 = commands2.Command()
    command3 = commands2.Command()

    start_spying_on(command1)
    start_spying_on(command2)
    start_spying_on(command3)

    command1.isFinished = lambda: True

    selectCommand = commands2.SelectCommand(
        {"one": command1, "two": command2, "three": command3}, lambda: "one"
    )

    scheduler.schedule(selectCommand)
    scheduler.run()

    verify(command1).initialize()
    verify(command1).execute()
    verify(command1).end(False)

    verify(command2, never()).initialize()
    verify(command2, never()).execute()
    verify(command2, never()).end(False)

    verify(command3, never()).initialize()
    verify(command3, never()).execute()
    verify(command3, never()).end(False)


def test_selectCommandInvalidKey(scheduler: commands2.CommandScheduler):
    command1 = commands2.Command()
    command2 = commands2.Command()
    command3 = commands2.Command()

    start_spying_on(command1)
    start_spying_on(command2)
    start_spying_on(command3)

    command1.isFinished = lambda: True

    selectCommand = commands2.SelectCommand(
        {"one": command1, "two": command2, "three": command3}, lambda: "four"
    )

    scheduler.schedule(selectCommand)


def test_selectCommandRequirement(scheduler: commands2.CommandScheduler):
    system1 = commands2.Subsystem()
    system2 = commands2.Subsystem()
    system3 = commands2.Subsystem()
    system4 = commands2.Subsystem()

    command1 = commands2.Command()
    command1.addRequirements(system1, system2)
    command2 = commands2.Command()
    command2.addRequirements(system3)
    command3 = commands2.Command()
    command3.addRequirements(system3, system4)

    start_spying_on(command1)
    start_spying_on(command2)
    start_spying_on(command3)

    selectCommand = commands2.SelectCommand(
        {"one": command1, "two": command2, "three": command3}, lambda: "one"
    )

    scheduler.schedule(selectCommand)
    scheduler.schedule(commands2.InstantCommand(lambda: None, system3))

    verify(command1).end(interrupted=True)
    verify(command2, never()).end(interrupted=True)
    verify(command3, never()).end(interrupted=True)



# validated: 2024-01-24 DS 192a28af4731 DeferredCommand.java
from typing import Callable

from wpiutil import SendableBuilder

from .command import Command
from .commandscheduler import CommandScheduler
from .printcommand import PrintCommand
from .subsystem import Subsystem


class DeferredCommand(Command):
    """
    Defers Command construction to runtime. Runs the command returned by the supplier when this
    command is initialized, and ends when it ends. Useful for performing runtime tasks before
    creating a new command. If this command is interrupted, it will cancel the command.

    Note that the supplier *must* create a new Command each call. For selecting one of a
    preallocated set of commands, use :class:`commands2.SelectCommand`.
    """

    def __init__(self, supplier: Callable[[], Command], *requirements: Subsystem):
        """
        :param supplier:     The command supplier
        :param requirements: The command requirements.
        """
        super().__init__()

        assert callable(supplier)

        self._null_command = PrintCommand(
            f"[DeferredCommand] Supplied command (from {supplier!r} was None!"
        )
        self._supplier = supplier
        self._command = self._null_command
        self.addRequirements(*requirements)

    def initialize(self):
        cmd = self._supplier()
        if cmd is not None:
            self._command = cmd
            CommandScheduler.getInstance().registerComposedCommands([self._command])
        self._command.initialize()

    def execute(self):
        self._command.execute()

    def isFinished(self):
        return self._command.isFinished()

    def end(self, interrupted):
        self._command.end(interrupted)
        self._command = self._null_command

    def initSendable(self, builder: SendableBuilder):
        super().initSendable(builder)
        builder.addStringProperty(
            "deferred",
            lambda: "null"
            if self._command is self._null_command
            else self._command.getName(),
            lambda _: None,
        )



# validated: 2024-01-20 DS 92aecab2ef05 button/CommandJoystick.java
from typing import Optional

from wpilib import Joystick
from wpilib.event import EventLoop

from ..commandscheduler import CommandScheduler
from .commandgenerichid import CommandGenericHID
from .trigger import Trigger


class CommandJoystick(CommandGenericHID):
    """
    A version of :class:`wpilib.Joystick` with :class:`.Trigger` factories for command-based.
    """

    _hid: Joystick

    def __init__(self, port: int):
        """
        Construct an instance of a controller.

        :param port: The port index on the Driver Station that the controller is plugged into.
        """

        super().__init__(port)
        self._hid = Joystick(port)

    def getHID(self) -> Joystick:
        """
        Get the underlying GenericHID object.

        :returns: the wrapped GenericHID object
        """
        return self._hid

    def trigger(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the trigger button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the trigger button's digital signal attached to the
            given loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getTrigger())

    def top(self, loop: Optional[EventLoop] = None) -> Trigger:
        """
        Constructs an event instance around the top button's digital signal.

        :param loop: the event loop instance to attach the event to, defaults
                     to :func:`commands2.CommandScheduler.getDefaultButtonLoop`

        :returns: an event instance representing the top button's digital signal attached to the given
                  loop.
        """
        if loop is None:
            loop = CommandScheduler.getInstance().getDefaultButtonLoop()
        return Trigger(loop, lambda: self._hid.getTop())

    def setXChannel(self, channel: int):
        """
        Set the channel associated with the X axis.

        :param channel: The channel to set the axis to.
        """
        self._hid.setXChannel(channel)

    def setYChannel(self, channel: int):
        """
        Set the channel associated with the Y axis.

        :param channel: The channel to set the axis to.
        """
        self._hid.setYChannel(channel)

    def setZChannel(self, channel: int):
        """
        Set the channel associated with the Z axis.

        :param channel: The channel to set the axis to.
        """
        self._hid.setZChannel(channel)

    def setThrottleChannel(self, channel: int):
        """
        Set the channel associated with the throttle axis.

        :param channel: The channel to set the axis to.
        """
        self._hid.setThrottleChannel(channel)

    def setTwistChannel(self, channel: int):
        """
        Set the channel associated with the twist axis.

        :param channel: The channel to set the axis to.
        """
        self._hid.setTwistChannel(channel)

    def getXChannel(self) -> int:
        """
        Get the channel currently associated with the X axis.

        :returns: The channel for the axis.
        """
        return self._hid.getXChannel()

    def getYChannel(self) -> int:
        """
        Get the channel currently associated with the Y axis.

        :returns: The channel for the axis.
        """
        return self._hid.getYChannel()

    def getZChannel(self) -> int:
        """
        Get the channel currently associated with the Z axis.

        :returns: The channel for the axis.
        """
        return self._hid.getZChannel()

    def getTwistChannel(self) -> int:
        """
        Get the channel currently associated with the twist axis.

        :returns: The channel for the axis.
        """
        return self._hid.getTwistChannel()

    def getThrottleChannel(self) -> int:
        """
        Get the channel currently associated with the throttle axis.

        :returns: The channel for the axis.
        """
        return self._hid.getThrottleChannel()

    def getX(self) -> float:
        """
        Get the x position of the HID.

        :returns: the x position
        """
        return self._hid.getX()

    def getY(self) -> float:
        """
        Get the y position of the HID.

        :returns: the y position
        """
        return self._hid.getY()

    def getZ(self) -> float:
        """
        Get the z position of the HID.

        :returns: the z position
        """
        return self._hid.getZ()

    def getTwist(self) -> float:
        """
        Get the twist value of the current joystick. This depends on the mapping of the joystick
        connected to the current port.

        :returns: The Twist value of the joystick.
        """
        return self._hid.getTwist()

    def getThrottle(self) -> float:
        """
        Get the throttle value of the current joystick. This depends on the mapping of the joystick
        connected to the current port.

        :returns: The Throttle value of the joystick.
        """
        return self._hid.getThrottle()

    def getMagnitude(self) -> float:
        """
        Get the magnitude of the direction vector formed by the joystick's current position relative to
        its origin.

        :returns: The magnitude of the direction vector
        """
        return self._hid.getMagnitude()

    def getDirectionRadians(self) -> float:
        """
        Get the direction of the vector formed by the joystick and its origin in radians.

        :returns: The direction of the vector in radians
        """
        return self._hid.getDirectionRadians()

    def getDirectionDegrees(self) -> float:
        """
        Get the direction of the vector formed by the joystick and its origin in degrees.

        :returns: The direction of the vector in degrees
        """
        return self._hid.getDirectionDegrees()
